from matplotlib import pyplot as plt
from lmz import *
import numpy as np
from scipy.optimize import curve_fit
import scipy.stats as stats
import matplotlib.pyplot as plt
import pandas as pd

import seaborn as sns


###############
# regression:
##################
def f(x, a, b,c):
    return a * np.log(x+c) +b

def regr(d):
    popt, pcov = curve_fit(f, Range(d) , d)
    return popt



###################
# READ THE DATA
###################
def readfile(fname):
    # format: 1 row contains all y for one x
    r=[]
    for line in open(fname,'r').read().split('\n'):
        values=line.split()
        r.append(np.array(Map(float,values)))
    return r


#################
# fit 
###############

def ex(d, out=2.576):
    """
        d: list of values(y) at a single x position
        
        return: [[outlier, low], 15.9 percentile, 50% percentile, 84.1 percentile, [outlier high]]
    """
    a,med,b = np.percentile(d, [15.9, 50, 84.1])
    oa = [v for v in d if v < med-out*(med-a)]
    ob = [v for v in d if v > med+out*(b-med)]

    return oa,a,med,b,ob



#################
# PUT STUFF TOGETHER 
###############

def draw(data,smooth = False):
    """
        data: list of list of numbers :) 

        does: plot median line, areafill+-sigma, outliers as +
    """
    stuff = [ex(rr,2.576) for rr in data]
    LR,LOW,MED,HI,HR = Transpose(stuff)
    X= Range(MED)


    p1,p2,p3 = [regr(x) for x in [LOW,MED,HI]]
    M2 =[f(x,*p2) for x in X]
    L2 =[f(x,*p1) for x in X]
    H2 =[f(x,*p3) for x in X]


    lo,med,hi = LOW, MED, HI 
    if smooth:
        lo,med,hi = L2,M2,H2

    # plot fillbetween
    plt.fill_between(X,lo,hi, color='green', alpha=0.4, label='median+-34.1th percentile')

    # plot median 
    plt.plot(X,med,label='median')

    # plot outlayers
    for i,(lr,hr) in enumerate(zip(LR,HR)):
        z=lr+hr
        if i==0:
            plt.scatter([i]*len(z), z,marker='x',c='b',alpha=.4,label='p<.99')
        else:
            plt.scatter([i]*len(z), z,marker='x',c='b',alpha=.4)

    plt.xticks(X)
    plt.legend()





def boxplot(d,x =1): 
    # how wide is the box? 
    width = min(.5,.15*(max(d)-min(d))) 

    # how tall is the box? 
    oa,a,med,b,ob = ex(d)
    hight = b-a

    # draw the box
    rectangle = plt.Rectangle((x-.5*width,a), width, hight,ec="black",fill=False)
    plt.gca().add_patch(rectangle)
    # draw the median 
    line = plt.Line2D((x-.5*width, x+.5*width), (med, med), lw=1.5, color='orange')
    plt.gca().add_line(line)
    # draw the outliers
    plt.scatter([x]*len(oa+ob), oa+ob, edgecolors='black', c='white')




def plotsmoothline(datax, datay):
    popt, pcov = curve_fit(f, datax ,datay)
    X = np.linspace(min(datax), max(datax), 100)
    plt.plot(X, map(lambda x:f(x,*popt), X))

if __name__ == "__main__":

    # import matplotlib
    # matplotlib.use("module://matplotlib-sixel")
    # import matplotlib.pyplot as plt

    data = readfile("../test/counts.txt")
    draw(data,False)
    plt.ylim((40,150))
    plt.show()

    boxplot([1,2,3,4,5,6,7,7,100])
    plt.show()




import numpy as np
import functools
from sklearn.mixture import GaussianMixture as gmm 
import multiprocessing as mp                                                
import math

###
# knee point detection 1: 
# choose value between gaussians
###

def gmm2(values,**kwargs):
    A=np.array(values).reshape(-1, 1)
    model = gmm(n_components=2, n_init=10,**kwargs).fit(A)
    return model.predict_proba(A), model

def only_between(cluster_probs,Y,gmm_means):
    '''removes outliers, i.e. sets 0 fields of z where a not between means'''
    mi,ma = gmm_means.min(),gmm_means.max()
    filtr = [ aa < mi or aa > ma for aa in Y ]
    cluster_probs[filtr]=0 
    return cluster_probs

def between_gaussians(values,**kwargs):
    cluster_probs, model = gmm2(values,**kwargs)
    cprobs = cluster_probs.sum(axis=1)
    cprobs = only_between(cprobs,values, model.means_)
    return np.argmax(cprobs)



####
#  max distance to line 
#  kneed implementaiton is annoying, 
######

def diag_maxdist(values, debug=True): 
    points = [ (x,y) for x,y in enumerate(values)  ]
    x1,y1 = points[0]
    x2,y2 = points[-1]
    def dist(p):
        x0,y0 = p
        return  ( (y2-y1)*x0 - (x2-x1)*y0 + x2*y1 -y2*x1) / math.sqrt(  (y2-y1)**2 + (x2-x1)**2 ) 


    res  =list(map(dist,points))
    if debug:
        if not (all ( map( lambda x: x>=0, res)) or all(map(lambda x: x<=0 ,res))):
            print ("ubergauss.diag_maxdist encountered strange data: dist to diagonal", res, points)
    return np.argmax(np.abs(res))






###
# multiprocessing
###
def mpmap(func, iterable, chunksize=1, poolsize=2):                            
    pool = mp.Pool(poolsize)                                                    
    result = pool.map(func, iterable, chunksize=chunksize)                      
    pool.close()                                                                
    pool.join()                                                                 
    return list(result) 


def traingmm(n_comp, X=None,n_init=10,**kwargs):
   return gmm(n_init = n_init,
     n_components=n_comp, **kwargs).fit(X) # covarianve_type full is default


####
# get a model
####

def get_model(X, poolsize = -1,
                 nclust_min = 4,
                 nclust_max = 20,
                 n_init = 30,
                 covariance_type = 'full',
                 use_bic = False, # bic or aic ;; aic should result in more clusters ;; penalty is lower for generating more clusters
                 kneepoint_detection = lambda x: diag_maxdist(x,False), 
                 debug = False,
                 **kwargs):

    # trivial case:
    if nclust_min == nclust_max: 
        return traingmm(nclust_min,X=X,n_init=n_init,**kwargs)

    # train models
    train = functools.partial(traingmm,X=X,n_init=n_init,**kwargs)
    if poolsize < 2: 
        models = [train(x) for x in range(nclust_min, nclust_max)]
    else:
        models = mpmap( train , range(nclust_min,nclust_max), poolsize= poolsize)

    # kneepoint
    scores = [m.bic(X) if use_bic else m.aic(X) for m in models]
    best = kneepoint_detection(scores)
    return models[best]




'''
####
# OLD MAN, using KNEED to find kneepoints... 
####
def  get_model(X, poolsize = 4, nclust_min = 4, nclust_max = 20, n_init = 20,**kwargs):
    import kneed
    # trivial case:
    if nclust_min == nclust_max: 
        return traingmm(nclust_min,X=X,n_init=n_init,**kwargs)

    # train models
    train = functools.partial(traingmm,X=X,n_init=n_init,**kwargs)
    models = mpmap( train , range(nclust_min,nclust_max), poolsize= poolsize)

    # kneepoint
    bics = [-m.bic(X) for m in models]
    print ("bics:", bics)
    #best = last_of_variate_gaussian(bics)
    kneedler = kneed.KneeLocator(list(range(len(bics))),bics, S=1.0, curve='concave', direction='increasing')
    best = kneedler.knee
    if not isinstance(best,int):
        best = 3
    return models[best]
'''

'''
#####
# knee point detection 2:
# choose last of the distribution that is more varying
########
def last_of_variate_gaussian(bics):
    predictions, model = gmm2(bics)
    zz = predictions[:,np.argmax(model.covariances_)]
    for i,v in enumerate(zz):
        if v < .95: 
            return i-1
    return -1
'''


